import os
import subprocess
import sys


def open_file(path):
    """
    Open given file
    """
    if sys.platform.startswith('linux'):
        ret_code = subprocess.call(['xdg-open', path])

    elif sys.platform.startswith('darwin'):
        ret_code = subprocess.call(['open', path])

    elif sys.platform.startswith('win'):
        ret_code = subprocess.call(['explorer', path])

    return ret_code


def open_folder(path):
    """
    Open given folder
    """

    if os.path.isfile(path):
        path = os.path.dirname(path)

    return open_file(path)



from .provider_base import ProviderBase


class ProviderMusicmatch(ProviderBase):
    def get_name(self):
        return "Musixmatch"

    def get_url(self, title, artist, album):
        return 'https://www.musixmatch.com/lyrics/{}/{}'.format(self.encode(artist),
                                                                self.encode(title))

    def get_xpath(self, title, artist, album):
        return '//p[@class="mxm-lyrics__content "]/text()'

    def get_regex(self, title, artist, album):
        return ''

    def get_node_separator(self):
        return '\n'

    def encode(self, param):
        param = (
            param
            .replace(' & ', '-')
            .replace('... ', '')
            .replace(' ', '-')
            .replace("'", '-')
            .replace('.', '')
            .replace('"', '')
            .replace(',', '')
            .replace('(', '')
            .replace('/', '-')
            .replace(')', '')
        )

        return super().encode(param)



import logging
import random
import re

import requests
from lxml import html


class ProviderBase(object):
    def get_lyrics(self, title, artist, album):
        try:
            url = self.get_url(title, artist, album)
            xpath = self.get_xpath(title, artist, album)
            regex = self.get_regex(title, artist, album)

            headers = {
                "User-Agent": self.get_user_agent(),
                "Accept": "*/*"
            }

            logging.info('Fetching ' + url)
            response = requests.get(url, headers=headers)

            if response.ok:

                nodes = html.fromstring(response.content).xpath(xpath)
                lyrics = self.get_node_separator().join(nodes)

                if regex:
                    match = re.search(regex, lyrics, re.IGNORECASE)
                    if match is not None:
                        lyrics = match.groups('')[0]

                lyrics = lyrics.strip()

                return lyrics

            else:
                return ''
        except:
            logging.exception('Could not fetch lyrics for ' + title)
            return ''

    def get_user_agent(self):
        return random.choice([
            "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; FSL 7.0.6.01001)",
            "Mozilla/5.0 (iPad; U; CPU OS 4_3_2 like Mac OS X; en-us) AppleWebKit/533.17.9 (KHTML, like Gecko) Version/5.0.2 Mobile/8H7 Safari",
            "Mozilla/5.0 (iPhone; CPU iPhone OS 5_0 like Mac OS X) AppleWebKit/534.46 (KHTML, like Gecko) Version/5.1 Mobile/9A334 Safari/7534.48.3",
            "Mozilla/5.0 (Linux; Android 4.4; Nexus 5 Build/_BuildID_) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/30.0.0.0 Mobile Safari/537.36",
            "Mozilla/5.0 (Linux; Android 5.1.1; Nexus 5 Build/LMY48B; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/43.0.2357.65 Mobile Safari/537.36",
            "Mozilla/5.0 (Linux; U; Android 4.1.1; en-gb; Build/KLP) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Safari/534.30",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_1) AppleWebKit/602.2.14 (KHTML, like Gecko) Version/10.0.1 Safari/602.2.14",
            "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
            "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:49.0) Gecko/20100101 Firefox/49.0",
            "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36",
        ])

    def get_name(self):
        pass

    def get_url(self, title, artist, album):
        pass

    def get_xpath(self, title, artist, album):
        pass

    def get_regex(self, title, artist, album):
        pass

    def get_node_separator(self):
        pass

    def encode(self, param):
        return requests.utils.quote(param)



import enum
import logging
import pathlib

import gi
gi.require_version('Gst', '1.0')
gi.require_version('GstAudio', '1.0')
from gi.repository import Gst, GstAudio


class Streamer(object):
    """ Audio engine, for playing music

    Attributes:
        _state: An enum {PLAYING, STOPED, PAUSED}
        _playbin: The gst element that can play a stream

    Properties:
        stream: The file or url that will play next
        state: An enum {PLAYING, STOPED, PAUSED}
        duration: A float representing the total duration of stream in seconds
        position: A float representing the position in seconds
        volume: A float (0.0 to 1.0) for the volume
    """

    class State(enum.Enum):
        PLAYING = enum.auto()
        STOPED = enum.auto()
        PAUSED = enum.auto()

    def __init__(self, about_to_finish=None, audio_changed=None, notify_volume=None):
        Gst.init(None)
        self.logger = logging.getLogger('Streamer')
        self._state = Streamer.State.STOPED
        self._playbin = Gst.ElementFactory.make('playbin', None)

        self.audio_changed = audio_changed

        if about_to_finish is not None:
            self._playbin.connect("about-to-finish", about_to_finish)

        if notify_volume is not None:
            self._playbin.connect('notify::volume', notify_volume)

    def play(self, stream=None):
        """Start the playback of the stream"""
        self.logger.debug('Play')
        self._playbin.set_state(Gst.State.READY)
        self.logger.debug('READY')

        if stream is not None:
            self.stream = stream

        self.resume()

    def stop(self):
        """Completly stop the playback of the stream"""
        self.logger.debug('Stop')
        self._playbin.set_state(Gst.State.NULL)
        self._state = Streamer.State.STOPED
        self.logger.debug('STOPED')

    def pause(self):
        """Temporaly stop the playback of the stream"""
        self.logger.debug('Pause')
        self._playbin.set_state(Gst.State.PAUSED)
        self._state = Streamer.State.PAUSED
        self.logger.debug('PAUSED')

    def resume(self):
        """Resume a temporaly stopped playback"""
        self.logger.debug('Resume')
        self._playbin.set_state(Gst.State.PLAYING)
        self._state = Streamer.State.PLAYING
        self.logger.debug('PLAYING')

    @property
    def stream(self):
        return self._playbin.props.uri

    @property
    def state(self):
        return self._state

    @property
    def duration(self):
        return self._playbin.query_duration(Gst.Format.TIME)[1] / Gst.SECOND

    @property
    def position(self):
        return self._playbin.query_position(Gst.Format.TIME)[1] / Gst.SECOND

    @property
    def volume(self):
        return self._playbin.get_volume(GstAudio.StreamVolumeFormat.CUBIC)

    @stream.setter
    def stream(self, value):
        if not value:
            self._playbin.props.uri = ''
        elif not value.startswith('file://'):
            self._playbin.props.uri = pathlib.Path(value).as_uri()
        else:
            self._playbin.props.uri = value

        self.logger.info(value)
        if self.audio_changed is not None:
            self.audio_changed()

    @volume.setter
    def volume(self, value):
        if value > 1:
            value = 1
        elif value < 0:
            value = 0

        self._playbin.set_volume(GstAudio.StreamVolumeFormat.CUBIC, value)

    @position.setter
    def position(self, value):
        self._playbin.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH, value * Gst.SECOND)



import enum
import peewee

from mokaplayer.core.database import Song


class AbstractPlaylist:
    """ Abstract class for a playlist
    """

    class OrderBy(enum.Enum):
        """ Enum for the different way to order a playlist
        """
        DEFAULT = enum.auto()
        ARTIST = enum.auto()
        ALBUM = enum.auto()
        TITLE = enum.auto()
        YEAR = enum.auto()
        LENGTH = enum.auto()
        ADDED = enum.auto()
        PLAYED = enum.auto()

    @property
    def name(self):
        """ Return the name of the playlist """

    def collections(self, order, desc):
        """ Return a list of songs in the specified order"""

    def get_orderby_fields(self, order, desc):
        """ Return a list of fields for a order by query
        """
        fields = []

        if order == self.OrderBy.ALBUM:
            fields = [peewee.fn.strip_articles(Song.Album),
                      Song.Discnumber, Song.Tracknumber]
        elif order == self.OrderBy.YEAR:
            fields = [Song.Year, peewee.fn.strip_articles(Song.Album),
                      Song.Discnumber, Song.Tracknumber]
        elif order == self.OrderBy.ADDED:
            fields = [-Song.Added, peewee.fn.strip_articles(Song.AlbumArtist), Song.Year,
                      peewee.fn.strip_articles(Song.Album), Song.Discnumber, Song.Tracknumber]
        elif order == self.OrderBy.TITLE:
            fields = [Song.Title]
        elif order == self.OrderBy.LENGTH:
            fields = [Song.Length]
        elif order == self.OrderBy.PLAYED:
            fields = [-Song.Played]
        else:
            fields = [peewee.fn.strip_articles(Song.AlbumArtist), Song.Year,
                      peewee.fn.strip_articles(Song.Album), Song.Discnumber, Song.Tracknumber]

        if desc:
            fields[0] = -fields[0]

        return fields


from .songs_playlist import SongsPlaylist
from .m3u_playlist import M3UPlaylist
from .mostplayed_playlist import MostPlayedPlaylist
from .rarelyplayed_playlist import RarelyPlayedPlaylist
from .recentlyadded_playlist import RecentlyAddedPlaylist
from .recentlyplayed_playlist import RecentlyPlayedPlaylist
from .upnext_playlist import UpNextPlaylist
from .albums_playlist import AlbumsPlaylist
from .album_playlist import AlbumPlaylist
from .artists_playlist import ArtistsPlaylist
from .artist_playlist import ArtistPlaylist



import logging
import pathlib
import uuid

import requests


class ProviderLastFM(object):
    def __init__(self, api_key, artwork_folder):
        if artwork_folder is None:
            raise TypeError('Missing required argument: artwork_folder')
        elif api_key is None:
            raise TypeError('Missing required argument: api_key')

        self.URL = 'http://ws.audioscrobbler.com/2.0/'
        self._api_key = api_key
        self._artwork_folder = artwork_folder

    def get_name(self):
        "LastFm"

    def get_album_artwork(self, album_name, artist_name):
        try:
            response = requests.get(self.URL, params={
                'method': 'album.getinfo',
                'artist': artist_name,
                'album': album_name,
                'autocorrect': '1',
                'api_key': self._api_key,
                'format': 'json'
            })

            json = response.json()

            if 'error' in json:
                logging.error('Could not fetch artwork for ' + album_name)
            else:
                image_url = json['album']['image'][:4][-1]['#text']
                if image_url:
                    filename = str(pathlib.Path(self._artwork_folder, str(uuid.uuid4())))
                    self._save_image_from_url(image_url, filename)
                    return filename

            return None

        except Exception as e:
            logging.exception('Could not fetch artwork for ' + album_name)
            return None

    def get_artist_artwork(self, artist_name):
        try:
            response = requests.get(self.URL, params={
                'method': 'artist.getinfo',
                'artist': artist_name,
                'api_key': self._api_key,
                'autocorrect': '1',
                'format': 'json'
            })

            json = response.json()

            if 'error' in json:
                logging.error('Could not fetch artwork for ' + artist_name)
                return None
            else:
                filename = str(pathlib.Path(self._artwork_folder, str(uuid.uuid4())))
                self._save_image_from_url(json['artist']['image'][2]['#text'], filename)
                return filename

        except:
            logging.exception('Could not fetch artwork for ' + artist_name)
            return None

    def _save_image_from_url(self, url, filename):
        try:
            response = requests.get(url)
            if response.ok:
                pathlib.Path(filename).parent.mkdir(exist_ok=True)

                with open(filename, 'wb') as f:
                    f.write(response.content)
                    return True
            else:
                logging.error('Could not download image from ' + url + 'response was not OK')
                return False
        except:
            logging.error('Could not download image from ' + url)
            return False



from mokaplayer.core.database import Song
from mokaplayer.core.playlists import AbstractPlaylist


class RarelyPlayedPlaylist(AbstractPlaylist):
    @property
    def name(self):
        return "Rarely Played"

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        field = Song.Played.desc() if desc else Song.Played.asc()
        return Song.select().where(Song.Played < 3).order_by(field).limit(70)



import logging
import mimetypes
import pathlib
import peewee
import time

from mokaplayer.core.database import (Album, Artist, Playlist, Song,
                                      database_context)
from mokaplayer.core import playlists
from mokaplayer.core.fetchers import artworks
from mokaplayer.core.m3u_parser import M3uParser


class Library(object):
    """Explore the music folder and extract songs, album, artist into the database

    Attributes:
        _musics_folder: A string indicating where the musics is located
        _playlist_folder: A string indicating where the playlists is located

    """

    INGORE_EXTENSION = {'.jpg', '.jpeg', '.db',
                        '.ini', '.png', '.bmp', '.pdf',
                        '.tif', '.txt', '.nfo', }

    def __init__(self, appconfig, userconfig):
        self.logger = logging.getLogger('Library')
        database_context.init(appconfig.DATABASE_FILE)

        Song.create_table(True)
        Album.create_table(True)
        Artist.create_table(True)
        Playlist.create_table(True)

        self.appconfig = appconfig
        self.userconfig = userconfig

    def is_musics_folder_valid(self):
        if not self.musics_folder:
            return False
        else:
            return pathlib.Path(self.musics_folder).is_dir()

    @property
    def musics_folder(self):
        return self.userconfig["library"]["music_directory"]

    @musics_folder.setter
    def musics_folder(self, value):
        self.logger.debug('Setting musics_folder: ' + value)
        self.userconfig["library"]["music_directory"] = value
        self.userconfig.save()

    @property
    def playlists_folder(self):
        folder = self.userconfig["library"]["playlist_directory"]
        if not folder or not pathlib.Path(folder).is_dir():
            folder = self.musics_folder
        return folder

    @playlists_folder.setter
    def playlists_folder(self, value):
        self.userconfig["library"]["playlist_directory"] = value
        self.userconfig.save()

    @property
    def artworks_folder(self):
        return self.appconfig.ARTWORK_CACHE_DIRECTORY

    def get_songs(self, paths):
        return Song.select().where(Song.Path << paths)

    def get_song(self, path):
        try:
            path = str(pathlib.Path(path).resolve(False))
            return Song.get(Song.Path == path)
        except:
            return None

    def get_playlist(self, name):
        try:
            x = Artist.get(Artist.Name == name)
            return playlists.ArtistPlaylist(x)
        except peewee.DoesNotExist:
            pass

        try:
            x = Album.get(Album.Name == name)
            return playlists.AlbumPlaylist(x)
        except peewee.DoesNotExist:
            pass

        return None

    def get_playlists(self):
        return [x.Path for x in Playlist.select(Playlist.Path)]

    def get_album(self, name, albumartist):
        try:
            return Album.get(Album.Name == name, peewee.fn.lower(Album.Artist) == albumartist.lower())
        except:
            return None

    def create_playlist(self, name):
        """Create a new playlist in the right folder"""

        path = pathlib.Path(self.playlists_folder) / (name + '.m3u')
        if not path.exists():
            M3uParser(str(path)).write()
            self.__sync_playlists()

    def sync_artwork(self):
        """ For every album and artist with a missing cover try to fetch it """
        with database_context.atomic():
            for album in Album.select():
                if not album.Cover or not pathlib.Path(album.Cover).exists():
                    path = pathlib.Path(album.Path).parent if pathlib.Path(album.Path).is_file else album.Path
                    album.Cover = artworks.get_album_artwork(self.appconfig.LASTFM_SECRET_API_KEY,
                                                             self.artworks_folder, album.Name,
                                                             album.Artist, path)
                    album.save()

            for artist in Artist.select():
                if not artist.Cover or not pathlib.Path(artist.Cover).exists():
                    artist.Cover = artworks.get_artist_artwork(self.appconfig.LASTFM_SECRET_API_KEY,
                                                               self.artworks_folder, artist.Name)
                    artist.save()

        self.logger.info(f'Artworks fetch completed')

    def sync(self):
        """Synchronize data from library and actual data in the musics folder

        Raises:
            ValueError: When musics_folder is not set
        """
        if self.musics_folder is None or not pathlib.Path(self.musics_folder).is_dir():
            raise ValueError('Invalid music folder: ' + str(self.musics_folder))

        self.logger.info(f"Scanning {self.musics_folder}")
        start = time.perf_counter()
        self.__sync_songs()
        self.__sync_artists()
        self.__sync_albums()
        self.__sync_playlists()
        end = time.perf_counter()
        self.logger.info('Scan ended in {:.3f}'.format(end - start))

    def __sync_songs(self):
        paths = []
        for x in pathlib.Path(self.musics_folder).glob('**/*'):
            if x.is_dir():
                continue
            if x.suffix.lower() in self.INGORE_EXTENSION:
                continue
            paths.append(str(x.resolve(False)))

        all_paths = set(paths)
        known_paths = {x.Path for x in Song.select(Song.Path)}
        new_paths = all_paths - known_paths
        deleted_paths = known_paths - all_paths

        with database_context.atomic():
            for index, path in enumerate(new_paths):
                mime = mimetypes.guess_type(path)
                if mime[0] and 'audio' in str(mime[0]) and 'mpegurl' not in str(mime[0]):
                    s = Song(Path=path)
                    s.read_tags()
                    s.save()
                if index % 300 == 0 and index > 0:
                    self.logger.info(f'Scanning songs {index}/{len(new_paths)}')

            for song in deleted_paths:
                Song.delete().where(Song.Path == song).execute()

        self.logger.info(f'Scanning songs completed')

    def __sync_playlists(self):
        self.logger.info('Scanning playlists')

        all_paths = set(str(x) for x in pathlib.Path(self.playlists_folder).glob('**/*m3u'))
        known_paths = {x.Path for x in Playlist.select(Playlist.Path)}
        new_paths = all_paths - known_paths
        deleted_paths = known_paths - all_paths
        with database_context.atomic():
            for path in new_paths:
                playlist = M3uParser(path)
                Playlist(Name=playlist.name, Path=playlist.location).save()
            for path in deleted_paths:
                Playlist.delete().where(Song.Path == path).execute()

    def __sync_artists(self):
        self.logger.info('Scanning artists')
        database_context.execute_sql("""
            INSERT INTO ARTIST ('Name')
            SELECT DISTINCT AlbumArtist FROM Song
            LEFT JOIN Artist ON Song.AlbumArtist = Artist.Name
            WHERE AlbumArtist != '' AND ArtistId IS NULL
        """)
        database_context.execute_sql("""
            DELETE FROM Artist
            where not exists (select songid
                              from song
                              where Song.AlbumArtist = Artist.Name)
        """)

    def __sync_albums(self):
        self.logger.info('Scanning albums')
        database_context.execute_sql("""
            INSERT INTO album ('Name', 'Year', 'Path', 'Artist')
            SELECT song.album, song.year, song.path, song.albumartist
            FROM   song
            LEFT JOIN album ON album.NAME = song.album AND album.artist LIKE song.albumartist
            WHERE  song.album != '' AND album.albumid IS NULL
            GROUP  BY song.album
        """)

        database_context.execute_sql("""
            DELETE FROM Album
            where not exists (select songid
                            from song
                            where album.NAME = song.album AND
                                    album.artist LIKE song.albumartist)
        """)

    # def vplayer_library_converter(self):
    #     import udatetime, json, gzip, os
    #     with DB.atomic():
    #         with gzip.open('/run/media/vincent/D-DRV/Documents/Autre/Dotfiles/VPlayer/library.gz', mode="rt") as f:
    #             json_library = json.loads(f.read())
    #             songs = json_library["songs"]
    #             for index, x in enumerate(songs):
    #                 for real in Song.select().where(Song.Path ** ('%' + x['path'][-20:])):
    #                     if os.path.samefile(real.Path, x['path']):
    #                         real.Added = udatetime.from_string(x.get('added', 0))
    #                         real.save()
    #                         break
    #                 if (index % 10 == 0):
    #                     print(f'{index}/{len(songs)}')



from .provider_lastfm import ProviderLastFM
from .provider_local_filesystem import ProviderLocalFileSystem


def get_album_artwork(lastfm_apikey, artwork_folder, album_name, artist_name, album_local_path):
    result = None

    if not result and album_local_path:
        result = ProviderLocalFileSystem().get_album_artwork(album_local_path)

    if not result and album_name and artist_name and lastfm_apikey:
        result = ProviderLastFM(lastfm_apikey, artwork_folder).get_album_artwork(album_name, artist_name)

    return result


def get_artist_artwork(lastfm_apikey, artwork_folder, artist_name):
    return ProviderLastFM(lastfm_apikey, artwork_folder).get_artist_artwork(artist_name)



class Event:
    def __init__(self):
        self.__handlers = []

    def subscribe(self, handler):
        self.__handlers.append(handler)

    def unsubscribe(self, handler):
        self.__handlers.remove(handler)

    def fire(self, *args, **keywargs):
        for handler in self.__handlers:
            handler(*args, **keywargs)

    def clear(self):
        self.__handlers.clear()



import os

import appdirs
from mokaplayer.config import secret


class Default:
    APPLICATION_NAME = 'MokaPlayer'
    USER_DIRECTORY = appdirs.user_data_dir(APPLICATION_NAME)
    CACHE_DIRECTORY = appdirs.user_cache_dir(APPLICATION_NAME)
    LASTFM_SECRET_API_KEY = secret.LASTFM_SECRET_API_KEY
    LOG_LEVEL = 'INFO'

    os.makedirs(USER_DIRECTORY, exist_ok=True)
    os.makedirs(CACHE_DIRECTORY, exist_ok=True)


class Developpement(Default):
    DATABASE_FILE = 'dev_database.db'
    USER_CONFIG_FILE = 'dev_config.yaml'
    PLAYER_CACHE_FILE = os.path.join(Default.CACHE_DIRECTORY, 'dev', 'player.gz')
    ARTWORK_CACHE_DIRECTORY = os.path.join(Default.CACHE_DIRECTORY, 'dev', 'artworks')
    LOG_LEVEL = 'DEBUG'


class Testing(Default):
    DATABASE_FILE = ':memory:'
    USER_CONFIG_FILE = 'test_config.yaml'
    PLAYER_CACHE_FILE = os.path.join(Default.CACHE_DIRECTORY, 'test', 'player.gz')
    ARTWORK_CACHE_DIRECTORY = os.path.join(Default.CACHE_DIRECTORY, 'test', 'artworks')
    LOG_LEVEL = 'INFO'


class Production(Default):
    DATABASE_FILE = os.path.join(Default.USER_DIRECTORY, 'database.db')
    USER_CONFIG_FILE = os.path.join(Default.USER_DIRECTORY, 'config.yaml')
    PLAYER_CACHE_FILE = os.path.join(Default.CACHE_DIRECTORY, 'player.gz')
    ARTWORK_CACHE_DIRECTORY = os.path.join(Default.CACHE_DIRECTORY, 'artworks')
    LOG_LEVEL = 'INFO'



import pkg_resources
from gi.repository import Gtk


class HelpShortcutsWindow():
    GLADE_FILE = pkg_resources.resource_filename('mokaplayer',
                                                 'data/ui/help_shortcuts_window.ui')

    @staticmethod
    def get_diaglog():
        builder = Gtk.Builder()
        builder.add_from_file(HelpShortcutsWindow.GLADE_FILE)
        return builder.get_object('shortcuts_window')



import collections
import random


class Queue(object):
    """ Hold an id of the song that is currently playing, and all the following one.

    Attributes:
        _container: A deque representing a list of objects in a specific order
        _current: Represent last object taken in the queue
    """

    def __init__(self):
        self._container = collections.deque()
        self._current = None

    def pop(self):
        """Remove the currently playing object"""
        tmp = self._current
        self._current = None
        return tmp

    def peek(self):
        """Get the currently playing object"""
        return self._current

    def next(self):
        """Move to the next object"""
        if any(self._container):
            if self._current == self._container[0]:
                self._container.append(self._container.popleft())
            self._current = self._container[0]
        else:
            self._current = None

    def prev(self):
        """Move to the next previous"""
        if any(self._container):
            if self._current == self._container[0]:
                self._container.appendleft(self._container.pop())
            self._current = self._container[0]
        else:
            self._current = None

    def seek(self, item):
        """Move to the item"""
        if item in self._container:
            while self._current != item:
                self.next()
        else:
            self.prepend(item)
            self.next()

    def shuffle(self):
        """Shuffle the container"""
        random.shuffle(self._container)
        try:
            self._container.remove(self._current)
            self._container.appendleft(self._current)
        except ValueError:
            pass

    def append(self, container):
        """Insert objects at the end of the container"""
        if not isinstance(container, collections.MutableSequence):
            container = [container]

        for x in container:
            try:
                self._container.remove(x)
            except ValueError:
                pass
            self._container.append(x)

        if self._current is None and any(self._container):
            self._current = self._container[0]

    def prepend(self, container):
        """Insert objects at the start of the container"""
        if not isinstance(container, collections.MutableSequence):
            container = [container]

        container.reverse()

        for x in container:
            if self._current != x:
                try:
                    self._container.remove(x)
                except ValueError:
                    pass
                self._container.appendleft(x)

        try:
            self._container.remove(self._current)
            self._container.appendleft(self._current)
        except ValueError:
            pass

        if self._current is None and any(self._container):
            self._current = self._container[0]

    def remove(self, container):
        """Remove objects from the container"""
        if not isinstance(container, collections.MutableSequence):
            container = [container]

        for x in container:
            try:
                self._container.remove(x)
            except ValueError:
                pass

    def clear(self):
        """Remove every objects from the container"""
        self._container.clear()

    def __len__(self):
        return len(self._container)

    def __iter__(self):
        return iter(self._container)



from gi.repository import Gtk
from mokaplayer.core.helpers import time_helper


class AdapterSongFromAlbum:
    @staticmethod
    def create_row(song):
        return [
            song.Path,
            str(song.Tracknumber) + ".",
            song.Title or 'Unknown',
            time_helper.seconds_to_string(song.Length)
        ]

    @staticmethod
    def create_store():
        model = Gtk.ListStore(str, str, str, str)
        model.set_default_sort_func(lambda *unused: 0)
        model.set_sort_column_id(
            Gtk.TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, Gtk.SortType.ASCENDING)
        return model

    @staticmethod
    def create_col_number():
        return [0, 1, 2, 3]



from mokaplayer.core.database import Song
from mokaplayer.core.playlists import AbstractPlaylist


class SongsPlaylist(AbstractPlaylist):

    @property
    def name(self):
        return "Songs"

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        if order == self.OrderBy.DEFAULT:
            order = self.OrderBy.ARTIST

        return Song.select().order_by(*self.get_orderby_fields(order, desc))



from gi.repository import Gtk
from . import AlbumView


class ArtistView:
    def __init__(self, artist_playlist, row_activated, button_press_event):
        self.artist_playlist = artist_playlist
        self.view = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.view.pack_start(Gtk.Separator(margin_bottom=20, margin_left=10, margin_right=10, margin_top=0), False, True, 0)

        for album_playlist in artist_playlist.collections_album():
            self.view.pack_start(AlbumView(album_playlist,
                                           row_activated,
                                           button_press_event).get_view(), False, False, 0)
        self.view.show_all()

    def get_view(self):
        return self.view



import pkg_resources
from gi.repository import Gtk


class InputBox(Gtk.Dialog):

    def __init__(self, parent, title, text):
        Gtk.Dialog.__init__(self, title, parent, 0,
                            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                             Gtk.STOCK_OK, Gtk.ResponseType.OK))

        self.set_default_size(350, 0)
        self.set_default_response(Gtk.ResponseType.OK)

        self.label = Gtk.Label(text)
        self.label.set_margin_start(5)
        self.label.set_margin_top(5)
        self.label.set_margin_bottom(5)
        self.label.set_margin_end(5)

        self.entry = Gtk.Entry()
        self.entry.set_margin_start(5)
        self.entry.set_margin_top(5)
        self.entry.set_margin_bottom(5)
        self.entry.set_margin_end(5)
        self.entry.connect('activate', lambda widget: self.response(Gtk.ResponseType.OK))

        self.get_content_area().add(self.label)
        self.get_content_area().add(self.entry)

    def get_text(self):
        self.show_all()
        response = self.run()
        text = self.entry.get_text()
        self.destroy()
        if response == Gtk.ResponseType.OK:
            return text
        else:
            return None



from mokaplayer.ui.gtk.windows.about_window import AboutWindow
from mokaplayer.ui.gtk.windows.help_shortcuts_window import HelpShortcutsWindow
from mokaplayer.ui.gtk.windows.inputbox_window import InputBox
from mokaplayer.ui.gtk.windows.lyrics_window import LyricsWindow
from mokaplayer.ui.gtk.windows.tabs_window import TabsWindow
from mokaplayer.ui.gtk.windows.tagseditor_window import TagsEditorWindow
from mokaplayer.ui.gtk.windows.main_window import MainWindow



import peewee

from mokaplayer.core.database import Artist
from mokaplayer.core.playlists import AbstractPlaylist


class ArtistsPlaylist(AbstractPlaylist):

    @property
    def name(self):
        return "Artists"

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False, search=''):
        if order == self.OrderBy.ARTIST or order == self.OrderBy.DEFAULT:
            fields = [peewee.fn.strip_articles(Artist.Name)]

        if desc:
            fields[0] = -fields[0]

        query = Artist.select()
        if search:
            query = query.where(Artist.Name.regexp(search))

        return query.order_by(*fields)



from mokaplayer.core.database import Album, Song
from mokaplayer.core.playlists import AbstractPlaylist


class AlbumPlaylist(AbstractPlaylist):

    def __init__(self, album=None, song_path=None):
        if album:
            self.album = album
        elif song_path:
            song = Song.get(Song.Path == song_path)
            self.album = Album.get(Album.Name == song.Album)

    @property
    def name(self):
        return f'{self.album.Name} ({self.album.Year})'

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        return Song.select().where(Song.Album == self.album.Name).order_by(Song.Discnumber, Song.Tracknumber)



import os
import pathlib


class ProviderLocalFileSystem(object):

    IMAGE_SUFFIXES = ['.JPG', '.JPEG', '.PNG', '.BMP']

    def __init__(self):
        pass

    def get_name(self):
        "Local files"

    def get_album_artwork(self, album_path):
        for file in sorted(pathlib.Path(album_path).glob('*'), key=os.path.getsize, reverse=True):
            if file.suffix.upper() in ProviderLocalFileSystem.IMAGE_SUFFIXES:
                return str(file.absolute())
        return None



import logging
import threading
import time

import pkg_resources
from gi.repository import Gdk, GObject, Gtk, Pango
from mokaplayer.core.helpers import time_helper
from mokaplayer.core.m3u_parser import M3uParser
from mokaplayer.core.database import Artist, Album
from mokaplayer.ui.gtk.adapter import AdapterSong
from mokaplayer.ui.gtk.helper import file_helper, image_helper
from mokaplayer.ui.gtk.controls import AlbumView, ArtistView
from mokaplayer.ui.gtk.windows import (AboutWindow, HelpShortcutsWindow,
                                       InputBox, LyricsWindow, TabsWindow,
                                       TagsEditorWindow)
from mokaplayer.core.playlists import (AbstractPlaylist, SongsPlaylist,
                                       M3UPlaylist, MostPlayedPlaylist,
                                       RarelyPlayedPlaylist,
                                       RecentlyAddedPlaylist,
                                       RecentlyPlayedPlaylist, UpNextPlaylist,
                                       AlbumsPlaylist, ArtistsPlaylist,
                                       AlbumPlaylist, ArtistPlaylist)


class MainWindow(Gtk.Window):
    """ Main window
    """

    GLADE_FILE = pkg_resources.resource_filename('mokaplayer', 'data/ui/main_window.ui')
    CSS_FILE = pkg_resources.resource_filename('mokaplayer', 'data/ui/application.css')
    ICON_FILE = pkg_resources.resource_filename('mokaplayer', 'data/icons/hicolor/48x48/apps/mokaplayer.png')

    def __init__(self, appconfig, userconfig, player):
        super().__init__(title="MokaPlayer", default_width=1366, default_height=768)

        self.logger = logging.getLogger('MainWindow')
        self.appconfig = appconfig
        self.userconfig = userconfig
        self.player = player
        self.current_playlist = SongsPlaylist()
        self.set_icon_from_file(self.ICON_FILE)
        self.has_flowbox_album_loaded = False
        self.has_flowbox_artist_loaded = False

        style_provider = Gtk.CssProvider()
        style_provider.load_from_path(self.CSS_FILE)
        Gtk.StyleContext.add_provider_for_screen(Gdk.Screen.get_default(), style_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        if self.userconfig['gtk']['darktheme']:
            settings = Gtk.Settings.get_default()
            settings.set_property("gtk-application-prefer-dark-theme", True)

        self.connect("destroy", self.on_window_destroy)
        self.connect("key-press-event", self.on_window_key_press)

        self.builder = Gtk.Builder()
        self.builder.add_from_file(self.GLADE_FILE)
        self.__get_object()
        self.__init_sort_radio()
        self.__init_gridview_columns()
        self.__init_sidebar()
        self.__init_txt_goto()

        self.builder.connect_signals(self)
        self.player.state_changed.subscribe(self.on_player_state_changed)
        self.player.audio_changed.subscribe(self.on_audio_changed)
        self.player.volume_changed.subscribe(self.on_volume_changed)

        self.__set_current_song_info()
        self.on_volume_changed()

        self.__show_current_playlist()

        GObject.timeout_add(750, self.on_tick, None)
        self.logger.info('Window loaded')

        if not self.player.library.is_musics_folder_valid():
            self.__ask_for_music_folder()

    def __ask_for_music_folder(self):
        dialog = Gtk.FileChooserDialog("Select where is your music", self,
                                       Gtk.FileChooserAction.SELECT_FOLDER,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        "Select", Gtk.ResponseType.OK))

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            self.player.library.musics_folder = dialog.get_filename()
            self.on_library_scan_activate(None)
        dialog.destroy()

    def __manage_goto(self):
        p = self.player.library.get_playlist(self.txt_goto.get_text())
        if p:
            self.current_playlist = p
            self.__show_current_playlist()
            self.goto_bar.set_search_mode(False)
            GObject.idle_add(lambda: self.txt_goto.set_text(''))

    def __init_txt_goto(self):
        completion = Gtk.EntryCompletion()
        store = Gtk.ListStore(str)

        artists_name = {x.Name for x in ArtistsPlaylist().collections()}
        albums_name = {x.Name for x in AlbumsPlaylist().collections()}

        for x in artists_name.union(albums_name):
            store.append([x])

        completion.set_model(store)
        completion.set_text_column(0)
        completion.set_inline_completion(True)
        completion.set_match_func(self.__txt_goto_match_func)
        completion.set_inline_selection(True)
        completion.connect('match_selected', lambda x, y, z: self.__manage_goto())
        self.txt_goto.set_completion(completion)

    def __txt_goto_match_func(self, completion, key_string, iter):
        modelstr = completion.get_model()[iter][0].lower()
        return key_string.lower() in modelstr

    def __init_gridview_columns(self):
        columns = self.gridview.get_columns()
        font = self.userconfig['grid']['font']
        for col in columns:
            col.set_visible(False)

            if font:
                for cell in col.get_cells():
                    cell.set_property('font', font)

        for name in reversed(self.userconfig['grid']['columns']):
            for col in columns:
                if name == col.get_title():
                    col.set_visible(True)
                    self.gridview.move_column_after(col, None)

    def __create_model(self):
        self.logger.info('Creating ListStore')
        start = time.perf_counter()

        model = AdapterSong.create_store()
        order = AbstractPlaylist.OrderBy[self.userconfig['grid']['sort']['field']]
        desc = self.userconfig['grid']['sort']['desc']
        songs = self.current_playlist.collections(order, desc)

        for row in songs:
            model.insert_with_valuesv(-1, AdapterSong.create_col_number(), AdapterSong.create_row(row))

        GObject.idle_add(lambda: self.__create_model_finished(model))

        end = time.perf_counter()
        self.logger.info('ListStore created in {:.3f} seconds'.format(end - start))

    def __create_model_finished(self, model):
        self.model = model.filter_new()
        self.model.set_visible_func(self.__model_filter_func)
        self.gridview.set_model(self.model)

    def __get_object(self):
        self.content = self.builder.get_object('content')
        self.gridview = self.builder.get_object('gridview')
        self.col_added = self.builder.get_object('col_added')
        self.col_length = self.builder.get_object('col_length')
        self.lbl_current_title = self.builder.get_object('lbl_current_title')
        self.lbl_current_song_infos = self.builder.get_object('lbl_current_song_infos')
        self.volume_scale = self.builder.get_object('volume_scale')
        self.btn_previous = self.builder.get_object('btn_previous')
        self.btn_play = self.builder.get_object('btn_play')
        self.btn_next = self.builder.get_object('btn_next')
        self.img_current_album = self.builder.get_object('img_current_album')
        self.img_play = self.builder.get_object('img_play')
        self.img_pause = self.builder.get_object('img_pause')
        self.txt_search = self.builder.get_object('txt_search')
        self.txt_goto = self.builder.get_object('txt_goto')
        self.lbl_current_time = self.builder.get_object('lbl_current_time')
        self.prb_current_time = self.builder.get_object('prb_current_time')
        self.radio_sort_artist = self.builder.get_object('radio_sort_artist')
        self.radio_sort_album = self.builder.get_object('radio_sort_album')
        self.radio_sort_title = self.builder.get_object('radio_sort_title')
        self.radio_sort_year = self.builder.get_object('radio_sort_year')
        self.radio_sort_length = self.builder.get_object('radio_sort_length')
        self.radio_sort_added = self.builder.get_object('radio_sort_added')
        self.radio_sort_played = self.builder.get_object('radio_sort_played')
        self.radio_sort_default = self.builder.get_object('radio_sort_default')
        self.chk_sort_desc = self.builder.get_object('chk_sort_desc')
        self.menu_gridview = self.builder.get_object('menu_gridview')
        self.search_bar = self.builder.get_object('search_bar')
        self.goto_bar = self.builder.get_object('goto_bar')
        self.menuchild_gridview_playlist = self.builder.get_object('menuchild_gridview_playlist')
        self.menu_gridview_remove_from_playlist = self.builder.get_object('menu_gridview_remove_from_playlist')
        self.listbox_playlist = self.builder.get_object('listbox_playlist')
        self.playlist_sidebar = self.builder.get_object('playlist_sidebar')
        self.lbl_playlist = self.builder.get_object('lbl_playlist')
        self.flowbox_artist = self.builder.get_object('flowbox_artist')
        self.flowbox_album = self.builder.get_object('flowbox_album')
        self.stack = self.builder.get_object('stack')
        self.album_viewport = self.builder.get_object('album_viewport')
        self.add(self.content)

    def __show_current_playlist(self):
        self.lbl_playlist.set_text(self.current_playlist.name)
        if isinstance(self.current_playlist, AlbumPlaylist):
            for child in self.album_viewport.get_children():
                self.album_viewport.remove(child)
            self.stack.set_visible_child_name('album_view')
            self.album_viewport.add(AlbumView(self.current_playlist,
                                              self.on_gridview_row_activated,
                                              self.on_gridview_button_press_event,
                                              hide_header=True).get_view())

        elif isinstance(self.current_playlist, ArtistPlaylist):
            for child in self.album_viewport.get_children():
                self.album_viewport.remove(child)
            self.stack.set_visible_child_name('album_view')
            self.album_viewport.add(ArtistView(self.current_playlist,
                                               self.on_gridview_row_activated,
                                               self.on_gridview_button_press_event).get_view())

        elif isinstance(self.current_playlist, AlbumsPlaylist):
            self.stack.set_visible_child_name('flowbox_album')
            self.flowbox_album.unselect_all()
            if not self.has_flowbox_album_loaded:
                self.__create_flowbox(self.flowbox_album)
                self.has_flowbox_album_loaded = True

        elif isinstance(self.current_playlist, ArtistsPlaylist):
            self.stack.set_visible_child_name('flowbox_artist')
            self.flowbox_artist.unselect_all()
            if not self.has_flowbox_artist_loaded:
                self.__create_flowbox(self.flowbox_artist)
                self.has_flowbox_artist_loaded = True
        else:
            self.gridview.set_model(AdapterSong.create_store())
            self.stack.set_visible_child_name('gridview')
            threading.Thread(target=self.__create_model).start()

    def __create_flowbox(self, flowbox):
        self.logger.info('Creating Flowbox')
        start = time.perf_counter()

        image_loading_queue = []
        children = []

        order = AbstractPlaylist.OrderBy[self.userconfig['grid']['sort']['field']]
        desc = self.userconfig['grid']['sort']['desc']
        image_size = self.userconfig['flowbox']['image_size']
        margin = self.userconfig['flowbox']['margin']
        collections = self.current_playlist.collections(order, desc, self.txt_search.get_text())

        for child in flowbox.get_children():
            flowbox.remove(child)

        for item in collections:
            image = Gtk.Image()
            image.set_size_request(image_size, image_size)
            image.get_style_context().add_class('cover')
            image_loading_queue.append((image, item.Cover, image_size, image_size))

            if isinstance(item, Album):
                children.append(self.__create_album_flowboxitem(item, image, margin))
            elif isinstance(item, Artist):
                children.append(self.__create_artist_flowboxitem(item, image, margin))

        for child in children:
            flowbox.add(child)
        flowbox.show_all()

        image_helper.set_multiple_image(image_loading_queue)

        end = time.perf_counter()
        self.logger.info('Flowbox created in {:.3f} seconds'.format(end - start))

    def __create_artist_flowboxitem(self, artist, image, margin):
        label_album = Gtk.Label(artist.Name, max_width_chars=0, justify=Gtk.Justification.LEFT,
                                ellipsize=Pango.EllipsizeMode.END, xalign=0, margin_top=5)
        label_album.get_style_context().add_class('text110')

        flowboxchild = Gtk.FlowBoxChild()
        grid = Gtk.Grid(margin=margin, halign=Gtk.Align.CENTER)
        grid.attach(image, 0, 0, 1, 1)
        grid.attach(label_album, 0, 1, 1, 1)
        flowboxchild.add(grid)
        flowboxchild.data = artist
        return flowboxchild

    def __create_album_flowboxitem(self, album, image, margin):
        label_album = Gtk.Label(f'{album.Name} ({album.Year})', max_width_chars=0,
                                justify=Gtk.Justification.LEFT, wrap=True,
                                wrap_mode=Pango.WrapMode.WORD_CHAR, xalign=0, margin_top=5)
        label_artist = Gtk.Label(album.Artist, max_width_chars=0, justify=Gtk.Justification.LEFT,
                                 ellipsize=Pango.EllipsizeMode.END, xalign=0, margin_top=5)

        label_album.get_style_context().add_class('text110')
        label_artist.get_style_context().add_class('dim-label')
        flowboxchild = Gtk.FlowBoxChild(halign=Gtk.Align.CENTER)
        grid = Gtk.Grid(margin=margin, halign=Gtk.Align.CENTER)
        grid.attach(image, 0, 0, 1, 1)
        grid.attach(label_album, 0, 1, 1, 1)
        grid.attach(label_artist, 0, 2, 1, 1)
        flowboxchild.add(grid)
        flowboxchild.data = album
        return flowboxchild

    def __create_sidebar_row(self, playlist):
        list_box_row = Gtk.ListBoxRow()
        list_box_row.playlist = playlist
        list_box_row.set_size_request(0, 40)
        list_box_row.add(Gtk.Label(playlist.name, margin_left=20, halign=Gtk.Align.START))
        return list_box_row

    def __create_sidebar_header(self, name):
        list_box_row = Gtk.ListBoxRow(activatable=False, selectable=False)
        list_box_row.set_size_request(0, 40)
        label = Gtk.Label(f'<b>{name}</b>', margin_left=10, halign=Gtk.Align.START, use_markup=True)
        list_box_row.add(label)
        return list_box_row

    def __create_playlist_menus(self):
        for child in self.menuchild_gridview_playlist.get_children():
            self.menuchild_gridview_playlist.remove(child)

        for child in self.listbox_playlist.get_children():
            self.listbox_playlist.remove(child)

        self.listbox_playlist.add(self.__create_sidebar_header('Library'))
        self.listbox_playlist.add(self.__create_sidebar_row(SongsPlaylist()))
        self.listbox_playlist.add(self.__create_sidebar_row(AlbumsPlaylist()))
        self.listbox_playlist.add(self.__create_sidebar_row(ArtistsPlaylist()))
        self.listbox_playlist.add(self.__create_sidebar_row(UpNextPlaylist(self.player.queue)))
        self.listbox_playlist.add(self.__create_sidebar_row(MostPlayedPlaylist()))
        self.listbox_playlist.add(self.__create_sidebar_row(RarelyPlayedPlaylist()))
        self.listbox_playlist.add(self.__create_sidebar_row(RecentlyPlayedPlaylist()))
        self.listbox_playlist.add(self.__create_sidebar_row(RecentlyAddedPlaylist()))
        self.listbox_playlist.add(self.__create_sidebar_header('Playlists'))

        for playlist_location in self.player.library.get_playlists():
            m3u = M3uParser(playlist_location)
            menu_item = Gtk.MenuItem(label=m3u.name)
            menu_item.connect('activate', self.on_menu_gridview_add_to_playlist_activate, m3u, self.menu_gridview)
            self.menuchild_gridview_playlist.append(menu_item)
            self.listbox_playlist.add(self.__create_sidebar_row(M3UPlaylist(m3u)))

        self.menuchild_gridview_playlist.show_all()
        self.listbox_playlist.show_all()

    def __set_current_song_info(self):
        self.logger.debug("Setting currrent song info")
        image_size = self.userconfig['header']['image_size']
        line1 = self.userconfig['header']['line1']
        line2 = self.userconfig['header']['line2']

        song = self.player.library.get_song(self.player.queue.peek())
        album = None

        if song:
            album = self.player.library.get_album(song.Album, song.AlbumArtist)
            self.lbl_current_title.set_text(line1.format(**song._data))
            self.lbl_current_song_infos.set_text(line2.format(**song._data))
        else:
            self.lbl_current_title.set_text('')
            self.lbl_current_song_infos.set_text('')

        args = (self.img_current_album, album.Cover if album else None, image_size, image_size)
        threading.Thread(target=image_helper.set_image, args=args).start()

    def __set_current_play_icon(self):
        if self.player.streamer.state == self.player.streamer.State.PLAYING:
            self.btn_play.set_image(self.img_pause)
        else:
            self.btn_play.set_image(self.img_play)

    def __model_filter_func(self, model, iter, data):
        return AdapterSong.search(self.txt_search.get_text(), model[iter])

    def __focus_song(self, path):
        for row in self.model:
            if row[0] == path:
                self.gridview.get_selection().select_path(row.path)
                self.gridview.set_cursor(row.path)
                self.gridview.scroll_to_cell(row.path, use_align=True)
                self.gridview.grab_focus()
                break

    def __show_lyrics(self, path):
        song = self.player.library.get_song(path)
        w = LyricsWindow()
        w.start_fetch(song)
        w.get_window().set_transient_for(self)
        w.get_window().show()

    def __show_tabs(self, path):
        song = self.player.library.get_song(path)
        w = TabsWindow()
        w.start_fetch(song)
        w.get_window().set_transient_for(self)
        w.get_window().show()

    def __show_tagseditor(self, paths):
        songs = list(self.player.library.get_songs(paths))
        w = TagsEditorWindow(songs)
        w.get_window().set_transient_for(self)
        w.get_window().show()

    def __library_scan(self):
        self.player.library.sync()
        GObject.idle_add(self.on_library_scan_finished)

    def __library_artwork(self):
        self.player.library.sync_artwork()
        GObject.idle_add(self.on_library_artworks_finished)

    def on_window_destroy(self, widget):
        self.player.save()
        Gtk.main_quit()

    def on_window_key_press(self, widget, event):
        keyval_name = Gdk.keyval_name(event.keyval)
        ctrl = (event.state & Gdk.ModifierType.CONTROL_MASK)

        if ctrl and keyval_name == 'f':
            search_mode = self.search_bar.get_search_mode()
            self.search_bar.set_search_mode(not search_mode)
            if not search_mode:
                self.txt_search.grab_focus()
                self.goto_bar.set_search_mode(False)
        elif ctrl and keyval_name == 'l':
            self.__show_lyrics(self.player.queue.peek())
        elif ctrl and keyval_name == 'p':
            search_mode = self.goto_bar.get_search_mode()
            self.goto_bar.set_search_mode(not search_mode)
            if not search_mode:
                self.txt_goto.grab_focus()
                self.search_bar.set_search_mode(False)
        elif ctrl and keyval_name == 'o':
            self.__focus_song(self.player.queue.peek())
        elif ctrl and keyval_name == 'g':
            if self.player.queue.peek():
                self.current_playlist = ArtistPlaylist(song_path=self.player.queue.peek())
                self.__show_current_playlist()
        elif ctrl and keyval_name == 'Left':
            self.player.streamer.volume -= 0.05
        elif ctrl and keyval_name == 'Right':
            self.player.streamer.volume += 0.05
        elif keyval_name == "Escape":
            self.txt_search.set_text('')
            self.txt_goto.set_text('')
            self.search_bar.set_search_mode(False)
            self.goto_bar.set_search_mode(False)
        elif keyval_name == "space" and not self.txt_search.has_focus():
            self.player.toggle()
        else:
            return False

        return True

    def on_listbox_playlist_row_activated(self, widget, row):
        self.current_playlist = row.playlist
        self.__show_current_playlist()

    def on_gridview_row_activated(self, widget, path, column):
        self.player.seek(widget.get_model()[path][0])

    def on_gridview_button_press_event(self, sender, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3:
            # if its a m3uPlaylist, offer option to remove selected songs
            self.menu_gridview_remove_from_playlist.set_visible(isinstance(self.current_playlist, M3UPlaylist))
            self.menu_gridview.popup(None, None, None, sender, event.button, event.time)
            self.menu_gridview.sender = sender
            return True

    def on_menu_gridview_append_activate(self, widget):
        treeview = widget.get_parent().sender
        self.player.queue.append(self.__get_selected_songs(treeview))

    def on_menu_gridview_insert_activate(self, widget):
        treeview = widget.get_parent().sender
        self.player.queue.prepend(self.__get_selected_songs(treeview))

    def on_menu_gridview_add_to_playlist_activate(self, widget, playlist, menu):
        treeview = menu.sender
        playlist.read()
        for path in self.__get_selected_songs(treeview):
            playlist.append(path)
        playlist.write()

    def on_menu_gridview_remove_from_playlist_activate(self, widget):
        treeview = widget.get_parent().sender
        if isinstance(self.current_playlist, M3UPlaylist):
            self.current_playlist.m3u_parser.read()
            for path in self.__get_selected_songs(treeview):
                self.current_playlist.m3u_parser.remove(path)
            self.current_playlist.m3u_parser.write()
            self.__show_current_playlist()

    def on_menu_gridview_replace_activate(self, widget):
        treeview = widget.get_parent().sender
        self.player.queue.clear()
        self.player.queue.append(self.__get_selected_songs(treeview))

    def on_menu_gridview_edit_activate(self, widget):
        treeview = widget.get_parent().sender
        selected_songs = self.__get_selected_songs(treeview)
        if any(selected_songs):
            self.__show_tagseditor(selected_songs)

    def on_menu_gridview_tabs_activate(self, widget):
        treeview = widget.get_parent().sender
        selected_songs = self.__get_selected_songs(treeview)
        if any(selected_songs):
            self.__show_tabs(selected_songs[0])

    def on_menu_gridview_lyrics_activate(self, widget):
        treeview = widget.get_parent().sender
        selected_songs = self.__get_selected_songs(treeview)
        if any(selected_songs):
            self.__show_lyrics(selected_songs[0])

    def on_menu_gridview_folder_activate(self, widget):
        treeview = widget.get_parent().sender
        selected_songs = self.__get_selected_songs(treeview)
        if any(selected_songs):
            file_helper.open_folder(selected_songs[0])

    def on_menu_gridview_goto_artist_activate(self, widget):
        treeview = widget.get_parent().sender
        selected_songs = self.__get_selected_songs(treeview)
        if any(selected_songs):
            self.current_playlist = ArtistPlaylist(None, song_path=selected_songs[0])
            self.__show_current_playlist()

    def on_menu_gridview_goto_album_activate(self, widget):
        treeview = widget.get_parent().sender
        selected_songs = self.__get_selected_songs(treeview)
        if any(selected_songs):
            self.current_playlist = AlbumPlaylist(None, song_path=selected_songs[0])
            self.__show_current_playlist()

    def __get_selected_songs(self, gridview):
        model, pathlist = gridview.get_selection().get_selected_rows()
        result = []
        for path in pathlist:
            result.append(model[path][0])
        return result

    def on_btn_next_clicked(self, widget):
        self.player.next()

    def on_btn_play_clicked(self, widget):
        self.player.toggle()
        self.__set_current_play_icon()

    def on_btn_previous_clicked(self, widget):
        self.player.prev()

    def on_audio_changed(self):
        GObject.idle_add(self.__set_current_song_info)
        GObject.idle_add(self.__set_current_play_icon)

    def on_player_state_changed(self):
        GObject.idle_add(self.__set_current_play_icon)

    def on_volume_changed(self):
        self.volume_scale.set_value(self.player.streamer.volume * 100)

    def on_volume_scale_value_changed(self, widget):
        self.player.streamer.volume = self.volume_scale.get_value() / 100

    def on_txt_goto_key_press_event(self, wiget, event):
        keyval_name = Gdk.keyval_name(event.keyval)
        if keyval_name == "Return":
            self.__manage_goto()

    def on_txt_search_key_press_event(self, widget, event):
        keyval_name = Gdk.keyval_name(event.keyval)
        if keyval_name == "Return" and any(self.model):
            self.__focus_song(self.model[0][0])

    def on_txt_search_search_changed(self, widget):
        self.model.refilter()
        text_len = len(self.txt_search.get_text())
        if text_len == 0 or text_len > 2:
            self.has_flowbox_artist_loaded = False
            self.has_flowbox_album_loaded = False
            if isinstance(self.current_playlist, ArtistsPlaylist):
                self.__show_current_playlist()
            elif isinstance(self.current_playlist, AlbumsPlaylist):
                self.__show_current_playlist()

    def on_player_open_stream_activate(self, event):
        pass

    def on_player_preferences_activate(self, event):
        pass

    def on_player_quit_activate(self, event):
        Gtk.main_quit()

    def on_library_scan_activate(self, event):
        threading.Thread(target=self.__library_scan).start()

    def on_library_scan_finished(self):
        self.__create_playlist_menus()
        self.has_flowbox_album_loaded = False
        self.has_flowbox_artist_loaded = False
        self.__show_current_playlist()

    def on_library_artworks_activate(self, event):
        threading.Thread(target=lambda: self.__library_artwork()).start()

    def on_library_artworks_finished(self):
        self.__set_current_song_info()
        self.has_flowbox_album_loaded = False
        self.has_flowbox_artist_loaded = False
        self.__show_current_playlist()

    def on_queue_add_activate(self, event):
        self.player.queue.clear()
        self.player.queue.append([x.Path for x in SongsPlaylist().collections()])
        self.player.queue.seek(self.player.queue.pop())

    def on_queue_add_playlist_activate(self, event):
        try:
            order = AbstractPlaylist.OrderBy[self.userconfig['grid']['sort']['field']]
            desc = self.userconfig['grid']['sort']['desc']
            self.player.queue.clear()
            self.player.queue.append([x.Path for x in self.current_playlist.collections(order, desc)])
        except AttributeError:
            pass  # Not all playlist contain songs

    def on_queue_clear_activate(self, event):
        self.player.queue.clear()

    def on_queue_shuffle_activate(self, event):
        self.player.queue.shuffle()

    def on_help_shortcuts_activate(self, event):
        shortcuts_window = HelpShortcutsWindow.get_diaglog()
        shortcuts_window.set_transient_for(self)
        shortcuts_window.show()

    def on_help_about_activate(self, event):
        about_window = AboutWindow.get_diaglog()
        about_window.set_transient_for(self)
        about_window.show()

    def on_playlist_create_activate(self, event):
        name = InputBox(self, "Playlist", 'What the name of the playlist?').get_text()
        if name:
            self.player.library.create_playlist(name)
            self.__create_playlist_menus()

    def on_view_toggle_search_activate(self, widget):
        self.search_bar.set_search_mode(not self.search_bar.get_search_mode())
        self.goto_bar.set_search_mode(False)

    def on_view_toggle_sidebar_activate(self, widget):
        self.__show_sidebar(not self.playlist_sidebar.get_reveal_child())

    def on_prp_current_time_click(self, widget, event):
        width = self.prb_current_time.get_allocated_width()
        duration = self.player.streamer.duration
        self.player.streamer.position = duration * event.x / width

    def on_tick(self, data):
        position = self.player.streamer.position
        duration = self.player.streamer.duration
        fraction = position / duration if duration else 0
        position_text = time_helper.seconds_to_string(position)
        duration_text = time_helper.seconds_to_string(duration)

        self.lbl_current_time.set_text(f'{position_text} / {duration_text}')
        self.prb_current_time.set_fraction(fraction)

        return True

    def __init_sidebar(self):
        show_sidebar = self.userconfig['sidebar']['show']
        self.__create_playlist_menus()
        if show_sidebar:
            self.playlist_sidebar.set_reveal_child(show_sidebar)

    def __show_sidebar(self, is_visible):
        self.playlist_sidebar.set_reveal_child(is_visible)
        self.userconfig['sidebar']['show'] = is_visible
        threading.Thread(target=self.userconfig.save).start()

    def __init_sort_radio(self):
        order = AbstractPlaylist.OrderBy[self.userconfig['grid']['sort']['field']]
        desc = self.userconfig['grid']['sort']['desc']

        self.chk_sort_desc.set_active(desc)
        if order == AbstractPlaylist.OrderBy.ARTIST:
            self.radio_sort_artist.set_active(True)
        elif order == AbstractPlaylist.OrderBy.ALBUM:
            self.radio_sort_album.set_active(True)
        elif order == AbstractPlaylist.OrderBy.TITLE:
            self.radio_sort_title.set_active(True)
        elif order == AbstractPlaylist.OrderBy.LENGTH:
            self.radio_sort_length.set_active(True)
        elif order == AbstractPlaylist.OrderBy.YEAR:
            self.radio_sort_year.set_active(True)
        elif order == AbstractPlaylist.OrderBy.ADDED:
            self.radio_sort_added.set_active(True)
        elif order == AbstractPlaylist.OrderBy.PLAYED:
            self.radio_sort_played.set_active(True)
        else:
            self.radio_sort_default.set_active(True)

    def on_sort_radio_toggled(self, widget):
        if widget is None or widget.get_active() or not isinstance(widget, Gtk.RadioButton):
            order = AbstractPlaylist.OrderBy.DEFAULT

            if self.radio_sort_default.get_active():
                order = AbstractPlaylist.OrderBy.DEFAULT
            elif self.radio_sort_artist.get_active():
                order = AbstractPlaylist.OrderBy.ARTIST
            elif self.radio_sort_album.get_active():
                order = AbstractPlaylist.OrderBy.ALBUM
            elif self.radio_sort_title.get_active():
                order = AbstractPlaylist.OrderBy.TITLE
            elif self.radio_sort_length.get_active():
                order = AbstractPlaylist.OrderBy.LENGTH
            elif self.radio_sort_year.get_active():
                order = AbstractPlaylist.OrderBy.YEAR
            elif self.radio_sort_added.get_active():
                order = AbstractPlaylist.OrderBy.ADDED
            elif self.radio_sort_played.get_active():
                order = AbstractPlaylist.OrderBy.PLAYED

            self.userconfig['grid']['sort']['desc'] = self.chk_sort_desc.get_active()
            self.userconfig['grid']['sort']['field'] = order.name

            threading.Thread(target=self.userconfig.save).start()
            self.has_flowbox_album_loaded = False
            self.has_flowbox_artist_loaded = False
            self.__show_current_playlist()

    def on_flowbox_selected_children_changed(self, widget):
        selected = widget.get_selected_children()
        if any(selected):
            if isinstance(selected[0].data, Album):
                self.current_playlist = AlbumPlaylist(selected[0].data)
            elif isinstance(selected[0].data, Artist):
                self.current_playlist = ArtistPlaylist(selected[0].data)

            self.__show_current_playlist()



import mokaplayer
import pkg_resources
from gi.repository import Gtk


class AboutWindow():
    GLADE_FILE = pkg_resources.resource_filename('mokaplayer',
                                                 'data/ui/about_window.ui')

    @staticmethod
    def get_diaglog():
        builder = Gtk.Builder()
        builder.add_from_file(AboutWindow.GLADE_FILE)
        dialog = builder.get_object('about_dialog')
        dialog.set_version(mokaplayer.__version__)

        return dialog



import logging
import os
import re

import requests
from lxml import html


class ProviderUltimateGuitar():

    SEARCH_URL = 'https://www.ultimate-guitar.com/search.php'
    DOWNLOAD_URL = 'https://tabs.ultimate-guitar.com/tab/download'

    @staticmethod
    def search(title, artist):
        """ Return a list of tabs (a dict with all infos) that can be fetched or downloaded"""

        response = requests.get(ProviderUltimateGuitar.SEARCH_URL,
                                params={
                                    'band_name': artist,
                                    'song_name': title,
                                    'type[]': ['200', '500'],  # tab and guitar pro
                                })

        if response.ok:
            try:
                page = html.fromstring(response.content)
                rows = page.xpath('//tr[.//a[contains(@class, "song") and contains(@class ,"result-link")]]')
                tabs = []

                for row in rows:
                    song = row.xpath('.//a[contains(@class, "song") and contains(@class ,"result-link")]')[0]
                    tab_type = row.xpath('./td[last()]//strong//text()')[0]
                    rating = row.xpath('.//b[contains(@class, "ratdig")]//text()')
                    rating = rating[0] if any(rating) else '0'

                    if tab_type == 'guitar pro' or tab_type == 'tab':
                        tabs.append({
                            'name': ' '.join(song.xpath('.//text()')).strip(),
                            'url': song.attrib['href'],
                            'rating': int(str(rating)),
                            'type': tab_type
                        })

                return sorted(tabs, key=lambda tab: (tab['type'],
                                                     -tab['rating']))
            except:
                logging.exception('Could not search guitar tabs for: ' + title)

        return []

    @staticmethod
    def fetch_ascii_tab(url):
        """Retrieve the ascii tab from a url"""
        try:
            response = requests.get(url)
            if response.ok:
                page = html.fromstring(response.content)
                nodes = page.xpath('//pre[contains(@class, "js-tab-content")]/text()')

                if any(nodes):
                    return nodes[0]
        except:
            logging.exception('Could not fetch ascii tabs for: ' + url)

        return ''

    @staticmethod
    def download_guitar_pro_tab(url, directory):
        """Retrieve and download the guitar pro tab (file) from a url"""
        try:
            response = requests.get(url)
            if response.ok:
                page = html.fromstring(response.content)
                nodes = page.xpath('//input[@id="tab_id"]/@value')

                if any(nodes):
                    try:
                        response = requests.get(ProviderUltimateGuitar.DOWNLOAD_URL,
                                                params={
                                                    'id': nodes[0]
                                                })
                        filename = re.findall('filename\s*?=\s?"(.+)"',
                                              response.headers['Content-Disposition'])[0]

                        os.makedirs(directory, exist_ok=True)
                        path = os.path.join(directory, filename)

                        with open(path, 'wb') as f:
                            f.write(response.content)
                            return path
                    except:
                        logging.exception('Could not download guitar pro tabs for id: ' + nodes[0])

        except:
            logging.exception('Could not download guitar pro tabs for: ' + url)

        return None



import peewee
from mokaplayer.core.database import Song
from mokaplayer.core.playlists import AbstractPlaylist


class RecentlyAddedPlaylist(AbstractPlaylist):
    @property
    def name(self):
        return "Recently Added"

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        field = [-Song.Added, peewee.fn.strip_articles(Song.AlbumArtist), Song.Year,
                 peewee.fn.strip_articles(Song.Album), Song.Discnumber, Song.Tracknumber]
        if desc:
            field[0] = Song.Added

        return Song.select().order_by(*field).limit(150)



import peewee
from playhouse.sqlite_ext import SqliteExtDatabase

database_context = SqliteExtDatabase(peewee.SqliteDatabase(None))


@database_context.func()
def strip_articles(text):
    articles = ['le ', 'la ', 'les ', 'the ', 'a ']
    for a in articles:
        if text.lower().startswith(a):
            return text[len(a):]

    return text


from .song import Song
from .artist import Artist
from .album import Album
from .playlist import Playlist



import datetime
import gzip
import json
import logging
import pathlib
import time
from threading import Thread

from mokaplayer.core.event import Event
from mokaplayer.core.keyboard import KeyboardClient
from mokaplayer.core.library import Library
from mokaplayer.core.queue import Queue
from mokaplayer.core.streamer import Streamer


class Player(object):

    def __init__(self, appconfig, userconfig):
        self.logger = logging.getLogger("Player")
        self.appconfig = appconfig
        self.userconfig = userconfig
        self.state_changed = Event()
        self.audio_changed = Event()
        self.volume_changed = Event()
        self.queue = Queue()
        self.streamer = Streamer(about_to_finish=self.__about_to_finish,
                                 audio_changed=self.__audio_changed,
                                 notify_volume=self.__notify_volume)

        self.library = Library(self.appconfig, self.userconfig)
        KeyboardClient(self)

    def play(self):
        self.logger.debug("Play")
        if self.streamer.state == Streamer.State.PAUSED:
            self.streamer.resume()
            self.state_changed.fire()
        elif self.streamer.state == Streamer.State.STOPED and self.queue.peek():
            self.streamer.play(self.queue.peek())
            self.state_changed.fire()

    def seek(self, path):
        self.logger.debug("Seek " + str(path))
        self.__set_play_count()
        self.queue.seek(path)
        self.streamer.play(path)

    def stop(self):
        self.logger.debug("Stop")
        self.__set_play_count()
        self.queue.pop()
        self.streamer.stop()
        self.state_changed.fire()
        self.audio_changed.fire()

    def pause(self):
        self.logger.debug("Pause")
        self.streamer.pause()
        self.state_changed.fire()

    def toggle(self):
        self.logger.debug("Toggle")
        if self.streamer.state == Streamer.State.PLAYING:
            self.streamer.pause()
            self.state_changed.fire()
        elif self.streamer.state == Streamer.State.PAUSED:
            self.streamer.resume()
            self.state_changed.fire()
        elif self.streamer.state == Streamer.State.STOPED and self.queue.peek():
            self.streamer.play(self.queue.peek())
            self.state_changed.fire()

    def next(self):
        self.logger.debug("Next")
        if len(self.queue):
            self.__set_play_count()
            self.queue.next()
            self.streamer.play(self.queue.peek())
        else:
            self.stop()

        self.state_changed.fire()

    def prev(self):
        self.logger.debug("Prev")
        if len(self.queue):
            self.__set_play_count()
            self.queue.prev()
            self.streamer.play(self.queue.peek())
        else:
            self.stop()

        self.state_changed.fire()

    def restore(self):
        try:
            self.logger.info("Restoring")
            if pathlib.Path(self.appconfig.PLAYER_CACHE_FILE).is_file():
                with gzip.open(self.appconfig.PLAYER_CACHE_FILE, 'rt') as file:
                    data = json.load(file)

                    self.streamer.volume = data['Volume']
                    self.queue.clear()
                    self.queue.append(data['Queue'])
                    self.seek(data['Current'])
                    self.pause()
                    time.sleep(0.3)
                    self.streamer.position = data['Position']
        except Exception as e:
            self.logger.exception('Could not restore player state')

    def save(self):
        try:
            self.logger.info("Saving")
            pathlib.Path(self.appconfig.PLAYER_CACHE_FILE).parent.mkdir(exist_ok=True)
            with gzip.open(self.appconfig.PLAYER_CACHE_FILE, 'wt') as file:
                json.dump({
                    "Queue": list(self.queue),
                    "Current": self.queue.peek(),
                    "Position": self.streamer.position,
                    "Volume": self.streamer.volume,
                }, file)
        except Exception as e:
            self.logger.exception('Could not save player state')

    def __about_to_finish(self, data):
        self.logger.debug("About to finish")
        self.__set_play_count()
        self.queue.next()
        self.streamer.stream = self.queue.peek()

    def __audio_changed(self):
        self.logger.debug("Audio changed")
        self.audio_changed.fire()

    def __notify_volume(self, data, data2):
        self.volume_changed.fire()

    def __set_play_count(self):
        song = self.library.get_song(self.queue.peek())
        try:
            percent = self.streamer.position / self.streamer.duration
        except ZeroDivisionError:
            percent = 0

        if song is not None and percent > 0.85:
            song.Played += 1
            song.Last_played = datetime.datetime.now()
            Thread(target=song.save).start()



import random
import requests

from .provider_azlyrics import ProviderAzLyrics
from .provider_darklyrics import ProviderDarkLyrics
from .provider_genius import ProviderGenius
from .provider_metrolyrics import ProviderMetrolyrics
from .provider_musixmatch import ProviderMusicmatch


def get(title, artist, album=''):
    providers = [
        ProviderAzLyrics(),
        ProviderDarkLyrics(),
        ProviderGenius(),
        ProviderMetrolyrics(),
        ProviderMusicmatch(),
    ]
    random.shuffle(providers)

    for provider in providers:
        lyrics = provider.get_lyrics(title, artist, album)

        if lyrics:
            return True, lyrics + "\n\nfrom " + provider.get_name()

    return False, ("Lyrics not found, try this: " +
                   "https://google.ca/search?q=" +
                   requests.utils.quote(f'{artist} {title} lyrics'))



from threading import Thread

import pkg_resources
from gi.repository import GObject, Gtk
from mokaplayer.core.fetchers import tabs as tabs_fetcher
from mokaplayer.ui.gtk.adapter import AdapterTab
from mokaplayer.ui.gtk.helper import file_helper


class TabsWindow():

    GLADE_FILE = pkg_resources.resource_filename('mokaplayer',
                                                 'data/ui/tabs_window.ui')

    def __init__(self):
        self.builder = Gtk.Builder()
        self.builder.add_from_file(self.GLADE_FILE)
        self.builder.connect_signals(self)
        self.window = self.builder.get_object('window')
        self.gridview = self.builder.get_object('gridview')
        self.textview = self.builder.get_object('textview')

    def start_fetch(self, song):
        self.song = song
        self.textview.get_buffer().set_text('')

        if self.song is not None:
            Thread(target=self.fetch).start()
        else:
            self.textview.get_buffer().set_text('')

    def fetch(self):
        result = tabs_fetcher.search(self.song.Title, self.song.Artist)
        model = AdapterTab.create_store()
        for tab in reversed(result):
            model.insert_with_valuesv(0,
                                      AdapterTab.create_col_number(),
                                      AdapterTab.create_row(tab))

        GObject.idle_add(lambda: self.on_fetch_finished(model))

    def on_fetch_finished(self, model):
        self.model = model
        self.gridview.set_model(model)

    def download(self, path):
        text = ''
        tab_type = self.model[path][0]
        name = self.model[path][1]
        url = self.model[path][3]

        if tab_type == 'tab':
            text = tabs_fetcher.fetch_ascii_tab(url)

        elif tab_type == 'guitar pro':
            path = tabs_fetcher.download_guitar_pro_tab(url, '/tmp/')

            if path is not None:
                text = f'{name} downloaded at "{path}"'
                file_helper.open_file(path)
            else:
                text = f'Error while downloading {name}'

        GObject.idle_add(lambda: self.on_download_finished(text))

    def on_download_finished(self, text):
        self.textview.get_buffer().set_text(text)

    def on_gridview_row_activated(self, treeview, path, column):
        Thread(target=lambda: self.download(path)).start()

    def get_window(self):
        return self.window



import peewee

from mokaplayer.core.database import Artist, Song, Album
from mokaplayer.core.playlists import AbstractPlaylist, AlbumPlaylist


class ArtistPlaylist(AbstractPlaylist):

    def __init__(self, artist=None, song_path=None):
        if artist:
            self.artist = artist
        elif song_path:
            song = Song.get(Song.Path == song_path)
            self.artist = Artist.get(Artist.Name == song.AlbumArtist)

    @property
    def name(self):
        return self.artist.Name

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        return (Song.select()
                .where(Song.AlbumArtist == self.artist.Name)
                .order_by(peewee.fn.strip_articles(Song.AlbumArtist), Song.Year,
                          peewee.fn.strip_articles(Song.Album), Song.Discnumber, Song.Tracknumber))

    def collections_album(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        return [AlbumPlaylist(x) for x in Album.select().where(peewee.fn.lower(Album.Artist) == self.artist.Name.lower()).order_by(Album.Year)]



import logging
import pathlib


class M3uParser:
    """Represent a list of path or url (media file) loaded from a M3U file

    Properties:
        location: A string representing the location of the playlist File
        name: A string representing the name of the file
    """

    def __init__(self, location):
        if not pathlib.Path(location).parent.is_dir():
            raise ValueError('Directory does not exist for: ' + location)

        self._location = location
        self._name = ''
        self._media_files = []

    def read(self):
        try:
            logging.debug('Reading playlist from ' + self.location)
            self.clear()
            with open(self.location, 'rt') as f:
                for line in f.read().split('\n'):
                    if line and not line.startswith('#EXT'):
                        self._media_files.append(line)
        except:
            logging.exception('Could not read playlist from ' + self.location)

    def write(self):
        """Write songs to the playlist file"""
        try:
            logging.debug('Writting playlist to ' + self.location)
            with open(self.location, 'wt') as f:
                f.write('#EXTM3U\n')
                for path in self._media_files:
                    f.write(path + '\n')
        except:
            logging.exception('Could not write playlist to ' + self.location)

    def clear(self):
        self._media_files.clear()

    def __len__(self):
        return len(self._media_files)

    def __iter__(self):
        self.read()
        return iter(self._media_files)

    def __getitem__(self, key):
        return self._media_files[key]

    def __setitem__(self, key, value):
        self._media_files[key] = value

    def __delitem__(self, key):
        del(self._media_files[key])

    def append(self, value):
        for file in self._media_files:
            if pathlib.Path(file).resolve() == pathlib.Path(value).resolve():
                return

        self._media_files.append(value)

    def remove(self, value):
        for file in list(self._media_files):
            if pathlib.Path(file).resolve() == pathlib.Path(value).resolve():
                self._media_files.remove(file)
                return

    @property
    def location(self):
        return self._location

    @property
    def name(self):
        if self._name:
            return self._name
        else:
            self._name = pathlib.Path(self.location).stem
            return self._name

    @name.setter
    def name(self, value):
        try:
            path = pathlib.Path(self.location)
            new_path = path.with_name(value + '.m3u')
            path.rename(new_path)
            self._name = value
            self._location = str(new_path)
        except:
            raise ValueError('Invalid name')



import pathlib
import pkg_resources

from multiprocessing.pool import ThreadPool

from gi.repository import GdkPixbuf
from gi.repository import GObject


def load(filename, width, height):
    if not filename or not pathlib.Path(filename).is_file():
        filename = pkg_resources.resource_filename('mokaplayer', 'data/placeholder.png')

    pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
    pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
    return pixbuf


def set_image(gtkimage, filename, width, height):
    pixbuf = load(filename, width, height)
    GObject.idle_add(lambda: gtkimage.set_from_pixbuf(pixbuf))


def set_multiple_image(list_tuples):
    pool = ThreadPool(4)
    for x in list_tuples:
        pool.apply_async(set_image, x)



def seconds_to_string(duration):
    """Convert a time in seconds to a mm:ss string
    """
    if duration is None:
        return '00:00'

    minutes = int(duration // 60)
    seconds = int(duration % 60)

    return '{:02d}:{:02d}'.format(minutes, seconds)



import logging


class KeyboardClient:
    def __init__(self, player):
        self.player = player
        self.logger = logging.getLogger('KeyboardClient')

        try:
            import gi
            gi.require_version('Keybinder', '3.0')
            from gi.repository import Keybinder
            Keybinder.init()
            Keybinder.bind('XF86AudioPlay', self._on_XF86AudioPlay)
            Keybinder.bind('XF86AudioNext', self._on_XF86AudioNext)
            Keybinder.bind('XF86AudioPrev', self._on_XF86AudioPrev)
        except ValueError:
            self.logger.warning('Keybinder is needed on Linux for MediaKey binding')

    def _on_XF86AudioNext(self, key):
        self.logger.info(key)
        self.player.next()

    def _on_XF86AudioPrev(self, key):
        self.logger.info(key)
        self.player.prev()

    def _on_XF86AudioPlay(self, key):
        self.logger.info(key)
        if self.player.streamer.state == self.player.streamer.State.PAUSED:
            self.player.play()
        elif self.player.streamer.state == self.player.streamer.State.PLAYING:
            self.player.pause()



import peewee

from mokaplayer.core.database import Album
from mokaplayer.core.playlists import AbstractPlaylist


class AlbumsPlaylist(AbstractPlaylist):

    @property
    def name(self):
        return "Albums"

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False, search=''):
        if order == self.OrderBy.ARTIST or order == self.OrderBy.DEFAULT:
            fields = [peewee.fn.strip_articles(Album.Artist), Album.Year]
        elif order == self.OrderBy.ALBUM:
            fields = [peewee.fn.strip_articles(Album.Name)]
        elif order == self.OrderBy.YEAR:
            fields = [Album.Year, peewee.fn.strip_articles(Album.Artist)]

        if desc:
            fields[0] = -fields[0]

        query = Album.select()
        if search:
            query = query.where(Album.Name.regexp(search) | Album.Artist.regexp(search))

        return query.order_by(*fields)



import pkg_resources

from gi.repository import Gtk

from mokaplayer.ui.gtk.helper import image_helper
from mokaplayer.ui.gtk.adapter.adapter_song_from_album import AdapterSongFromAlbum


class AlbumView:
    GLADE_FILE = pkg_resources.resource_filename('mokaplayer', 'data/ui/album_view.ui')

    def __init__(self, album_playlist, row_activated, button_press_event, hide_header=False):
        self.album_playlist = album_playlist
        self.builder = Gtk.Builder()
        self.builder.add_from_file(self.GLADE_FILE)

        self.view = self.builder.get_object("view")
        self.grid_songs = self.builder.get_object("grid_songs")
        self.img_cover = self.builder.get_object("img_cover")
        self.lbl_album_name = self.builder.get_object("lbl_album_name")
        self.lbl_year = self.builder.get_object("lbl_year")
        self.header = self.builder.get_object("header")

        self.grid_songs.connect('row-activated', row_activated)
        self.grid_songs.connect('button-press-event', button_press_event)
        self.grid_songs.connect('focus-in-event', self.on_grid_songs_focus_in_event)

        if hide_header:
            self.view.remove(self.header)
        else:
            self.lbl_album_name.set_text(self.album_playlist.album.Name)
            self.lbl_year.set_text(self.album_playlist.album.Year)

        image_helper.set_image(self.img_cover, album_playlist.album.Cover, 220, 220)

        self._build_grid_songs()

    def get_view(self):
        return self.view

    def on_grid_songs_focus_in_event(self, widget, event):
        """ We don't want 2 treeview with a selection,
            so we walk trougth the siblings and their children
            then call unselect_all() if it's a treeview
        """
        stack = [self.get_view().get_parent()]
        while any(stack):
            element = stack.pop()
            if isinstance(element, Gtk.TreeView) and element is not widget:
                element.get_selection().unselect_all()
            elif isinstance(element, Gtk.Container):
                stack.extend(element.get_children())

    def _build_grid_songs(self):
        model = AdapterSongFromAlbum.create_store()

        for song in self.album_playlist.collections():
            model.insert_with_valuesv(-1,
                                      AdapterSongFromAlbum.create_col_number(),
                                      AdapterSongFromAlbum.create_row(song))
        self.grid_songs.set_model(model)



import peewee
from mokaplayer.core.database import database_context


class Playlist(peewee.Model):
    PlaylistId = peewee.PrimaryKeyField()
    Name = peewee.CharField(index=True)
    Path = peewee.CharField()

    class Meta:
        database = database_context



from mokaplayer.core.database import Song
from mokaplayer.core.playlists import AbstractPlaylist


class RecentlyPlayedPlaylist(AbstractPlaylist):
    @property
    def name(self):
        return "Recently Played"

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        field = Song.Last_played.asc() if desc else Song.Last_played.desc()
        return Song.select().order_by(field).limit(25)



import pathlib

from mokaplayer.core.database import Song
from mokaplayer.core.playlists import AbstractPlaylist


class M3UPlaylist(AbstractPlaylist):
    def __init__(self, m3u_parser):
        self.m3u_parser = m3u_parser

    @property
    def name(self):
        return self.m3u_parser.name

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        paths = list(self.m3u_parser) if not desc else list(reversed(self.m3u_parser))
        playlist_folder = pathlib.Path(self.m3u_parser.location).parent
        query = Song.select().where(Song.Path << paths |
                                    Song.Path << [playlist_folder / (path) for path in paths])

        if order == self.OrderBy.DEFAULT:
            result = {x.Path: x for x in query}
            for path in paths:
                if path in result:
                    yield result[path]
                else:
                    path_abs = str(playlist_folder / (path))
                    if path_abs in result:
                        yield result[path_abs]
        else:
            for song in (query.order_by(*self.get_orderby_fields(order, desc))):
                yield song



from mokaplayer.core.database import Song
from mokaplayer.core.playlists import AbstractPlaylist


class MostPlayedPlaylist(AbstractPlaylist):
    @property
    def name(self):
        return "Most Played"

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        field = Song.Played.asc() if desc else Song.Played.desc()
        return Song.select().where(Song.Played > 0).order_by(field).limit(70)



from mokaplayer.core.database import Song
from mokaplayer.core.playlists import AbstractPlaylist


class UpNextPlaylist(AbstractPlaylist):
    def __init__(self, queue):
        self.queue = queue

    @property
    def name(self):
        return "Up next"

    def collections(self, order=AbstractPlaylist.OrderBy.DEFAULT, desc=False):
        result = {x.Path: x for x in Song.select()}
        for path in list(self.queue):
            if not path == self.queue.peek():
                yield result[path]
